from itertools import accumulate as _accumulate
from operator import add as _add

import numpy as _np


def sorting(_les_seq, _les, dim):
    l = [(_les[i], _les_seq[:, i]) for i in range(dim)]
    return _np.array([s[1] for s in l]).T, _np.array([s[0] for s in l])


def calc_les_differential(differential, **options):

    if differential().jacobian() is None:
        return calc_les_differential_w_orth(differential, **options)

    else:
        return calc_les_differential_w_qr(differential, **options)


def calc_les_differential_w_qr(differential,
                h=0.001, N=100000, u0=None, n_average=100,
                dynamic_para=None,
                error_func=False,
                dynamic_func=False,
                **options):

    model = _make_model(differential, h=h, u0=u0, **options)

    if error_func or dynamic_func:
        jacobian = model.dynamic_jacobian
        func_solve = lambda t, u, **option : \
            model.solve(t, u, f=model.dynamic_eq, **option)
    else:
        jacobian = model.j
        func_solve = model.solve

    # main
    dim = model.dim

    Q = _np.eye(dim)
    dp = dynamic_para or [dict() for _ in range(N)]

    les = []
    for _, p in enumerate(dp):
        U = (_np.eye(dim)+jacobian(**p)*h)@Q
        Q, R = _np.linalg.qr(U)

        les.append(_np.log(_np.abs(R.diagonal())))

        func_solve(model.t, model.u, h=h, **p)

    les_list = [l/(i*h) for i, l in enumerate(_accumulate(les, _add), 1)]
    les_average = _np.average(les_list[-n_average:], axis=0)
    return sorting(_np.array(les_list), les_average, dim)


def calc_max_le_differential(differential,
                     h=0.01, N=10000, u0=None, n_average=100, n_split=10,
                     dynamic_para=None,
                     error_func=False,
                     dynamic_func=False,
                     **options):

    model = _make_model(differential, u0=u0, h=h, **options)

    if error_func  or dynamic_func:
        func_solve = lambda t, u, **option : \
            model.solve(t, u, f=model.dynamic_eq, **option)
    else:
        func_solve = model.solve

    # main
    w_s = _np.random.random(model.dim)*(10**(-3))
    z_0 = _np.linalg.norm(w_s)

    u_tilde = model.u
    u = u_tilde + w_s

    u_tilde_list = [u_tilde]

    dp = dynamic_para or [dict() for _ in range(N)]

    for i, para in enumerate(dp):
        _, u_tilde = func_solve(i*h, u_tilde, h=h, **para)
        u_tilde_list.append(u_tilde)

    w_e = w_s
    v = _np.log(_np.linalg.norm(w_e)*(1/z_0))
    T = 1/n_split
    les_list = [v/T]

    for i, u_tilde in enumerate(u_tilde_list[n_split::n_split]):
        for j in range(n_split):
            d = dp[i*n_split+j]
            _, u = func_solve(i*n_split*h+j*h, u, h=h, **d)

        w_e = u - u_tilde
        v += _np.log(_np.linalg.norm(w_e)*(1/z_0))

        T=(1/n_split)*(i+2)
        les_list.append(v/T)

        w_s = z_0/_np.linalg.norm(w_e)*w_e
        u = u_tilde+w_s

    return les_list, _np.average(les_list[-n_average:])


def calc_les_differential_w_orth(differential,
                  h=0.01, N=10000, u0=None, n_average=100,
                  dynamic_para=None,
                  error_func=False,
                  dynamic_func=False,
                  **options):

    model = _make_model(differential, u0=u0, h=h, **options)

    if error_func or dynamic_func:
        func_solve = lambda t, u, **option : \
            model.solve(t, u, f=model.dynamic_eq, **option)
    else:
        func_solve = model.solve

    dim = model.dim

    # main
    d_0 = _np.array([_np.random.rand(dim) for _ in range(dim)])
    d_0 = _np.linalg.qr(d_0.T)[0].T

    u_tilde = model.u
    u_hat = [u_tilde + w for w in d_0]

    lm = [[] for _ in range(dim)]

    dp = dynamic_para or [dict() for _ in range(N)]

    for i, para in enumerate(dp):
        t = i*h
        _, u_tilde = func_solve(t, u_tilde, **para)

        d_tau = []
        for i in range(dim):
            _, u_hat[i] = func_solve(t, u_hat[i], h=h, **para)
            d_tau.append(u_hat[i] - u_tilde)
        d_tau = _np.array(d_tau)

        d_tau_bot = _np.linalg.qr(d_tau.T)[0].T

        for i in range(dim):
            on = _np.linalg.norm(d_0[i])
            d_0[i] =  on * d_tau_bot[i]
            u_hat[i] = u_tilde + d_0[i]

        d_ups, d_downs = _np.ones(3), _np.ones(3)
        for i in range(dim):
            d_ups = _np.linalg.norm(_np.outer(d_ups, d_tau[i]))
            d_downs = _np.linalg.norm(_np.outer(d_downs, d_0[i]))

            lm[i].append(_np.log(d_ups / d_downs))

    calc_l = lambda l : [sum(l[:i])/(i*h) for i in range(1, len(l)+1)]
    les_plus_list = [_np.array(calc_l(l)) for l in lm]

    les_list = []
    les_list.append(les_plus_list[0])
    for i in range(1, dim):
        les_list.append(les_plus_list[i] - les_plus_list[i-1])
    les_average = [_np.average(l[-n_average:]) for l in les_list]

    les_average=sorted(les_average, reverse=True)
    return sorting(_np.array(les_list).T, les_average, dim)


def _make_model(differential, u0=None, h=0.01, **options):
    if u0 is None:
        model = differential.on_attractor(h=h, **options)
    else:
        model = differential(**options)
        model.u, model.t = u0, 0
    return model



from ._systems import DynamicalSystems as _DynamicalSystems
from ._tu import TU as _TU


class Difference(_DynamicalSystems):

    def _solve(self, t, u, *, h=1, **params):
        self.u = self(t, u, **params)
        self.t = t+h

        return _TU(self.t, self.u)



from ..systems._difference import Difference as _Difference

class Logistic(_Difference):

    def parameter(self, a=4.0):
        self.a = a
        self.dim = 1

    def equation(self, t, u):
        return self.a * u * (1 - u)

    def jacobian(self):
        j = self.a*(1-2*self.u)
        return j



from itertools import cycle as _cycle

import matplotlib as _mpl
import numpy as _np

from ._utils import (bartlett as _bartlett,
                     get_minidx_below_seq as _get_minidx_below_seq,
                     reshape as _reshape)
from ..utils import Drawing as _Drawing

def _autocovariance_function(dim, u_seq, u_bar, tau, N):

    value = _np.zeros(dim)

    for t in range(N-1-tau):
        value+= (u_seq[t]-u_bar)*(u_seq[t+tau]-u_bar)

    return value/(N-tau)


def _autocorrelation_function(dim, u_seq, u_bar, tau, N):
    gamma_seq = _autocovariance_function(dim, u_seq, u_bar, tau, N)
    return gamma_seq/gamma_seq[0]


def acf(u_seq, tau):
    u_seq = _reshape(u_seq)

    N, dim = u_seq.shape

    u_bar = _np.average(u_seq, axis=0)

    gamma_seq = _np.array([_autocovariance_function(dim, u_seq, u_bar, t, N)
                           for t in range(tau)])

    return gamma_seq/gamma_seq[0]


def est_lag_w_acf(u_seq, tau_max, alpha=0.95,
                  plot=True, path_save_plot=None):

    rho_seq = acf(u_seq, tau_max)
    var_seq = _bartlett(rho_seq, alpha)
    idx_list = _get_minidx_below_seq(rho_seq, var_seq)

    if plot:
        color = _cycle(_mpl.rcParams['axes.prop_cycle'])
        d = _Drawing()
        for i, idx in enumerate(idx_list):
            c = next(color)['color']
            d[0,0].plot(var_seq[:, i],
                        label=f'standard error - {i}', color=c, linestyle='dashed')
            d[0,0].stem(rho_seq[:, i], label=f'acf - {i}',
                        markerfmt='.', linefmt=c, use_line_collection=True)
            d[0,0].scatter(idx, rho_seq[idx, i],
                           zorder=10, marker='o', color='red', s=20)

        d[0,0].legend()
        d[0,0].set_axis_label('Time~lag', 'Correlation~coefficient')
        d[0,0].set_xlim(0, tau_max-1)
        if path_save_plot is not None:
            d.save(path_save_plot)
        d.show()
        d.close()

    return idx_list



"""
% Using any of these subpackages requires an explicit import.
equations
utils
"""

from .__version__ import __version__

from .systems import *
from .exploration import *
from .exploration._utils import embedding
from .lyapunov import calc_les, calc_lyapunov_dimension
from .utils._draw import Drawing



import numpy as _np

from ._systems import DynamicalSystems as _DynamicalSystems
from ._solver import RK4 as _RK4
from ._tu import TU as _TU


class Differential(_DynamicalSystems):

    def _solve(self, t, u, *, h=0.01, solver=_RK4, f=None, **params):
        sol = solver(f or self, h=h, **params)
        self.u = sol(t, _np.array(u))
        self.t = t+sol.h

        return _TU(self.t, self.u)



from itertools import combinations as _combinations
from typing import NamedTuple as _NamedTuple

import numpy as _np

from ..utils import Drawing as _Drawing


class _DimensionResult(_NamedTuple):
    dimension : float  # 次元
    idx_dimension : int  # 最も相関係数が高かった時のインデックス

    min_correlation : float #
    correlations : _np.ndarray  # 相関係数
    slopes : _np.ndarray  # 1次関数によるフィッテイング時の傾き
    intercepts: _np.ndarray  # 1次関数によるフィッテイング時の切片

    log_frac_1_eps : _np.ndarray
    log_Ns : _np.ndarray
    dimensions : _np.ndarray  # 定義に沿った容量次元


class CalcDimension(object):

    def __init__(self, u_seq, scale_down, depsilon, base, loop,
                 batch_ave, min_correlation, plot, path_save_plot):
        self.dim, self.u_seq = self.check_dim(u_seq)
        if scale_down:
            self.u_seq = _scale_down(self.u_seq)

        self.batch_ave = batch_ave
        self.plot = plot
        self.min_correlation = min_correlation
        self.config_accuracy = (depsilon, base, loop)
        self.path_save_plot = path_save_plot

        self.length = len(self.u_seq)

    def __call__(self, q=0):
        result = self.main()

        if self.plot:
            _plot(result, self.batch_ave, self.path_save_plot, q=q)

        return result.dimension

    def main(self):
        accuracies, values = self.calc()

        log_accuracies = self.wrap_accuracies_in_main(accuracies)
        new_values =  self.wrap_value_in_main(values)
        dimensions = new_values / log_accuracies

        correlations, slopes, intercepts = \
            self._get_correlations_and_slopes(new_values, log_accuracies)
        self.correlations = correlations
        self.slopes = slopes

        idx = self._decide_idx_ref_mode(correlations)

        dimension = self.decide_dimension(idx, dimensions)

        return _DimensionResult(float(dimension), idx, self.min_correlation,
                                correlations, slopes, intercepts,
                                log_accuracies, new_values, dimensions)

    def calc(self):
        '''
        accuracyごとに計算(func)を行う.
        '''
        accuracies = self.make_accuracies(*self.config_accuracy)
        value_list = [self.func(epsilon) for epsilon in accuracies]
        return accuracies, _np.array(value_list)

    def func(self, epsilon):
        '''
        value = func(epsilon)
        '''

        x = self.u_seq[:, 0]
        xedges = self.make_edges(x, epsilon)

        if self.dim == 1:
            value = self.func_for_1dim(x, xedges)

        elif self.dim == 2:
            y = self.u_seq[:, 1]
            yedges = self.make_edges(y, epsilon)
            value = self.func_for_2dim(x, y, xedges, yedges)

        elif self.dim == 3:
            y, z = self.u_seq[:, 1], self.u_seq[:, 2]
            yedges = self.make_edges(y, epsilon)
            zedges = self.make_edges(z, epsilon)
            value_list = []
            for z_left in zedges:
                z_right = z_left + epsilon
                new_u_seq = self.u_seq[(z_left<z) & (z<=z_right)]
                new_x, new_y = new_u_seq[:, 0], new_u_seq[:, 1]
                value_list.append(self.func_for_2dim(new_x, new_y, xedges, yedges))
            value = self.wrap_value_3dim(value_list)

        else:
            value = 0

        return self.wrap_value(value)

    def func_for_1dim(self, x, xedges):
        return 0

    def func_for_2dim(self, x, y, xedges, yedges):
        return 0

    def decide_dimension(self, idx, dimensions):
        return _np.average(dimensions[idx:idx+self.batch_ave])

    def _decide_idx_ref_mode(self, correlations):
        correlations_over = _np.where(
            correlations>=self.min_correlation, correlations, 0)

        idx = int(_np.argmax(correlations_over))

        return idx

    def _get_correlations_and_slopes(self, h_seq, v_seq):
        batch_ave = self.batch_ave
        correlation_list, slope_list, intercept_list = [], [], []

        for i in range(len(h_seq)-batch_ave):
            h_seq_batch, v_seq_batch = h_seq[i:i+batch_ave], v_seq[i:i+batch_ave]

            correlation = _np.corrcoef(h_seq_batch, v_seq_batch)[0, 1]
            correlation_list.append(correlation)

            slope_now, intercept = _np.polyfit(v_seq_batch, h_seq_batch, 1)
            slope_list.append(slope_now)
            intercept_list.append(intercept)

        correlations = _np.array(correlation_list)
        slopes = _np.array(slope_list)
        intercepts = _np.array(intercept_list)

        return correlations, slopes, intercepts

    @staticmethod
    def wrap_value(value):
        return value

    @staticmethod
    def wrap_accuracies_in_main(accuracies):
        return _np.log(1/accuracies)

    @staticmethod
    def wrap_value_3dim(value_list):
        return sum(value_list)

    @staticmethod
    def wrap_value_in_main(values):
        return _np.log(values)

    @staticmethod
    def make_edges(a, ep):
        return _np.arange(_np.min(a)-ep, _np.max(a)+2*ep, ep)

    @staticmethod
    def check_dim(u_seq):
        if len(u_seq.shape)==1:
            u_seq = u_seq.reshape(len(u_seq), 1)
        return u_seq.shape[1], u_seq

    @staticmethod
    def make_accuracies(depsilon=0.02, base=7, loop=250):
        '''
        eベースでのaccuracyを作成する. 刻み幅はdepsilonとし,
        最小はe^(-base), 最大はe^((loop-1)*depsilon-base)となる.
        デフォルトではe^(-2)からe^(-7)のaccuracyのリストを返す.
        '''
        epsilon_list = [_np.e**(i*depsilon-base) for i in range(loop+1)]
        return _np.array(epsilon_list)[::-1]


class Capacity(CalcDimension):

    def func_for_1dim(self, x, xedges):
        H, _ = _np.histogram(x, bins=xedges)
        return _np.sum(H>0)

    def func_for_2dim(self, x, y, xedges, yedges):
        H, _, _ =  _np.histogram2d(x, y, bins=(xedges, yedges))
        return _np.sum(H>0)


class Information(CalcDimension):

    def func_for_1dim(self, x, xedges):
        H, _ = _np.histogram(x, bins=xedges)
        p = H/self.length
        return p[p>0]

    def func_for_2dim(self, x, y, xedges, yedges):
        H, _, _ =  _np.histogram2d(x, y, bins=(xedges, yedges))
        p = H/self.length
        return p[p>0]

    @staticmethod
    def wrap_value(p):
        return -1*_np.sum(_np.multiply(p, _np.log2(p)))

    @staticmethod
    def wrap_value_3dim(value_list):
        return _np.concatenate(value_list)

    @staticmethod
    def wrap_value_in_main(values):
        return values

    @staticmethod
    def wrap_accuracies_in_main(accuracies):
        return _np.log2(1/accuracies)


class Correlation(CalcDimension):
    '''
    Grassberger-Procaccia Algorithm (グラスバーガー - プロカッチャ アルゴリズム)
    '''
    def __init__(self, *args, **kwargs):
        super(Correlation, self).__init__(*args, **kwargs)
        self.distance = _np.array(
            [_dist(x_i, x_j) for x_i, x_j in _combinations(self.u_seq, 2)])

    def calc(self):
        '''
        accuracyごとに計算(func)を行う.
        '''
        accuracies = self.make_accuracies(*self.config_accuracy)
        crs = 2*_np.array(
            [_correlation_integrals(r, self.distance, len(self.u_seq))
             for r in accuracies])

        return accuracies, _np.array(crs)

    def decide_dimension(self, idx, dimensions):
        return _np.average(self.slopes[self.min_correlation<=self.correlations])

    @staticmethod
    def wrap_value_in_main(values):
        return -_np.log(values)

def _dist(a, b):
    return _np.linalg.norm(a - b)


def _correlation_integrals(r, distance, N):
    return _np.sum(r > distance)/(N**2)


def calc_dimension_capacity(u_seq, depsilon=0.02, base=7, loop=250,
                            min_correlation=0.999, scale_down=True,
                            batch_ave=10,  plot=True, path_save_plot=None):

    capacity = Capacity(
        u_seq, scale_down=scale_down,
        depsilon=depsilon, base=base, loop=loop,
        batch_ave=batch_ave, min_correlation=min_correlation,
        plot=plot, path_save_plot=path_save_plot)

    return capacity()


def calc_dimension_information(u_seq, depsilon=0.02, base=7, loop=250,
                               min_correlation=0.999, scale_down=True,
                               batch_ave=10,  plot=True, path_save_plot=None):

    infomation = Information(
        u_seq, scale_down=scale_down,
        depsilon=depsilon, base=base, loop=loop,
        batch_ave=batch_ave, min_correlation=min_correlation,
        plot=plot, path_save_plot=path_save_plot)

    return infomation(q=1)


def calc_dimension_correlation(u_seq, depsilon=0.02, base=7, loop=250,
                               min_correlation=0.999, scale_down=True,
                               batch_ave=10,  plot=True, path_save_plot=None):

    correlation = Correlation(
        u_seq, scale_down=scale_down,
        depsilon=depsilon, base=base, loop=loop,
        batch_ave=batch_ave, min_correlation=min_correlation,
        plot=plot, path_save_plot=path_save_plot)

    return correlation(q=2)


def _plot(result, batch, path_save_plot, q):
    color = {'b':'tab:blue', 'o':'tab:orange', 'g':'tab:green'}
    label_vertical_list = [r'$\ln N(\epsilon)$', r'$H(\epsilon)$',
                           r'$- \ln C(\epsilon)$']
    label_eq_list = [
        r'$\frac{\ln {N(\epsilon)}}{\ln \frac{1}{\epsilon}}$',
        r'$\frac{ {H}}{\log_2 \frac{1}{\epsilon}}$',
        r'$\frac{-\ln {C(\epsilon)}}{\ln \frac{1}{\epsilon}}$']

    x = result.log_frac_1_eps
    log_Ns = result.log_Ns
    idx = result.idx_dimension
    a = result.dimension

    def poly(_x):
        b = log_Ns[idx] - a*x[idx]
        return _np.poly1d((a, b))(_x)

    slopes = result.slopes
    dimensions = result.dimensions
    correlations = result.correlations
    min_correlation = result.min_correlation

    x_lr = [x[0], x[-1]]
    y = poly(x_lr)

    s_batch = slice(idx, idx+batch)

    d = _Drawing(3, 2, number=True, figsize=(3.14*1.7*2, 3.14*2),
                number_place=(0.96, 0.04), number_size=10)

    label_D = f'$D_{q}$'

    d[0,0].scatter(x, log_Ns, s=3, color=color['b'])
    d[0,0].scatter(x[s_batch], log_Ns[s_batch], s=3, color=color['o'])
    d[0,0].plot(x_lr, y, color=color['o'], linewidth=0.5)

    if q != 2:
        d[1,0].scatter(x, dimensions, s=3, color=color['b'],
                       label=label_eq_list[q])

    d[1,0].scatter(x[:-batch], slopes, s=3, color='tab:green', label='slope')
    d[1,0].axhline(a, color=color['o'], linewidth=0.5, linestyle='dashed')

    d[2,0].scatter(x[:-batch], correlations, s=3, color=color['b'])
    d[2,0].scatter(x[idx],correlations[idx], color=color['o'])

    d[2,0].axhline(1, color='black', linestyle='dashed',linewidth=0.5)
    d[2,0].axhline(min_correlation, color='black', linestyle='dashed',
                   linewidth=0.5)

    d[0,0].set_ylabel(label_vertical_list[q])
    d[1,0].set_ylabel(label_D)
    d[1,0].legend(loc='lower left')
    d[2,0].set_ylim(min_correlation-(1-min_correlation)/10, 1.0001)
    if q != 1:
        d[2,0].set_axis_label(r'\ln \frac{1}{\epsilon}', r'correlation')
    else:
        d[2,0].set_axis_label(r'\log_2 \frac{1}{\epsilon}', r'correlation')
    for i in range(3):
        d[i,0].set_xlim(min(x), max(x))

    if idx<batch:
        s = slice(idx, idx+batch*2)
        s2 = slice(0, idx+batch*2)
    elif idx>len(correlations)-2*batch:
        s = slice(idx-batch, len(correlations))
        s2 = slice(idx-batch, len(x))
    else:
        s = s2 = slice(idx-batch, idx+batch*2)

    x_lr = [x[s2][0], x[s2][-1]]

    d[0,1].scatter(x[s2], log_Ns[s2])
    d[0,1].scatter(x[s_batch], log_Ns[s_batch])
    d[0,1].plot(x_lr, poly(x_lr), color=color['o'])

    if q!=2:
        d[1,1].scatter(x[s2], dimensions[s2], color=color['b'])
    d[1,1].scatter(x[s], slopes[s], color='tab:green')
    d[1,1].axhline(a, color=color['o'], linestyle='dashed')

    d[2,1].scatter(x[s], correlations[s])
    d[2,1].scatter(x[idx], correlations[idx])
    d[2,1].axhline(1, color='black', linestyle='dashed')
    d[2,1].axhline(min_correlation, color='black', linestyle='dashed')


    d[0,1].set_ylabel(label_vertical_list[q])
    d[1,1].set_ylabel(label_D)
    if q!=1:
        d[2,1].set_axis_label(r'\ln \frac{1}{\epsilon}', r'correlation')
    else:
        d[2,1].set_axis_label(r'\log_2 \frac{1}{\epsilon}', r'correlation')
    d[2,1].set_ylim(min_correlation-(1-min_correlation), 1.0001)
    for i in range(3):
        d[i,1].set_xlim(min(x_lr), max(x_lr))

    if path_save_plot is not None:
        d.save(path_save_plot)

    d.show()
    d.close()


def _scale_down(seq):
    v_max = seq.max(axis=0, keepdims=True)
    v_min = seq.min(axis=0, keepdims=True)
    return seq/_np.max(v_max-v_min)



from abc import ABC as _ABC, abstractmethod as _abstractmethod
from collections import deque as _deque
from inspect import getfullargspec as _getfullargspec
from itertools import cycle as _cycle

import numpy as _np


class Synchronization(_ABC):

    _signal_seq = None

    @_abstractmethod
    def signal(self):
        """sending signal"""

    @_abstractmethod
    def leading(self):
        '''leading model'''

    @_abstractmethod
    def supporting(self):
        '''supporting model'''

    def set_model(self, shift=0):
        dq = _deque([self.leading, self.supporting])
        dq.rotate(shift)
        self._cyc =_cycle(dq)
        return self._cyc

    def _sync(self):
        return next(self._cyc)

    def get_signal(self):
        if self._signal_seq is None:
            self._signal_seq = []

        signal = self.signal()
        self._signal_seq.append(signal)
        return signal

    @property
    def signal_seq(self):
        return _np.array(self._signal_seq)

    @property
    def sync(self):
        try:
            return next(self._cyc)
        except AttributeError:
            raise AttributeError(f'need {self.__class__.__name__}.set_model()')


def _make_parameter(func, params):
    signals = _getfullargspec(func).args
    params = {s:p for s, p in zip(signals[3:], params)}
    params['f'] = func
    return params


def _check_oscillator(oscillators):
    for o in oscillators:
        if not isinstance(o, Synchronization):
            raise TypeError(
                f'{o.__class__.__name__} need to have Synchronization')


def coupling_oneway(o1, o2, N, **params):
    _check_oscillator([o1, o2])

    for _ in range(N):
        signal1 = o1.get_signal()
        signal2 = o2.get_signal()

        params1 = _make_parameter(o1.leading, [signal2])
        params2 = _make_parameter(o2.supporting, [signal1])

        o1.solve(*o1.internal_state, **dict(params1, **params))
        o2.solve(*o2.internal_state, **dict(params2, **params))

    return o1, o2



from ..systems._difference import Difference as _Difference

class Henon(_Difference):

    def parameter(self, a=1.4, b=0.3):
        self.a, self.b = a, b
        self.dim = 2

    def equation(self, t, u):
        x, y = u

        x_next = 1-self.a*x**2 + y
        y_next = self.b*x

        return x_next, y_next

    def jacobian(self):
        x, _ = self.u

        j = [[-self.a*2*x, 1],
             [self.b,      0]]

        return j




# Grassberger-Procaccia Algorithm (グラスバーガー - プロカッチャ アルゴリズム)

from itertools import combinations as _combinations
import warnings as _warnings

import numpy as _np


def _dist(a, b):
    return _np.linalg.norm(a - b)


def _correlation_integrals(r, distance, N):
    return _np.sum(r > distance)/(N**2)


def _calc_cr(e_seq, base_r, h_r, loop):
    distance = _np.array(
        [_dist(x_i, x_j) for x_i, x_j in _combinations(e_seq, 2)])

    rs = _np.array([_np.e**(i*h_r-base_r) for i in range(loop)])[::-1]
    crs = 2*_np.array(
        [_correlation_integrals(r, distance, len(e_seq)) for r in rs])
    return rs, crs


def calc_correlation_dimention_w_gp(e_seq,
                                    base=8, h_r=0.05, loop=200, batch_ave=10,
                                    normalize=True):
    """
    bは相関係数を計算するためのバッチサイズです。
    プロットの直線部分が相関係数0.999以上の場合の最大の勾配をD_2としている。
    """
    msg = 'It will not be available after version 0.2. ' \
          'Use `calc_dimension_correlation` instead.'
    _warnings.warn(msg)

    if normalize:
        v_max = e_seq.max(axis=0, keepdims=True)
        v_min = e_seq.min(axis=0, keepdims=True)
        e_seq = (e_seq-v_min)/(v_max-v_min)

    rs, crs = _calc_cr(e_seq, base, h_r, loop)

    correlation_max, slope = 0, None
    for i in range(len(crs)-batch_ave):
        log_cr, log_r = _np.log(crs[i:i+batch_ave]), _np.log(rs[i:i+batch_ave])
        correlation = _np.corrcoef(log_cr, log_r)[0,1]
        if (correlation >= 0.999) and (correlation > correlation_max):
            slope, _ = _np.polyfit(log_r, log_cr ,1)
            correlation_max = correlation

    return slope, rs, crs



from itertools import accumulate as _accumulate
from operator import add as _add

import numpy as _np


def _calc_lyapunov_dimension(les):
    sum_le = _np.array(list(_accumulate(les, _add)))
    ok = les[sum_le>0]
    d = len(ok)

    return d + sum(ok)/_np.abs(les[d])


def calc_lyapunov_dimension(les):
    if not isinstance(les, _np.ndarray):
        les = _np.array(les)
    return _calc_lyapunov_dimension(les)


if __name__ == '__main__':
    les = [0.943, 0.020, -14.667]
    D_L = calc_lyapunov_dimension(les)
    print(D_L)



from functools import partial as _partial

import numpy as _np


class Solver(object):

    A = NotImplemented
    B = NotImplemented
    n_stages = NotImplemented

    def __init__(self, f, *, h=0.01, **params):
        self.f, self.h = f, h
        self.params = params

    def __call__(self, t, u, **params):
        params = params or self.params
        K = _np.zeros((self.n_stages, u.shape[0]), dtype=u.dtype)
        return rk_step(self.f, t, u, self.A, self.B, K, self.h, **params)

    def __repr__(self):
        v = vars(self)
        f_name = v['f'].__name__
        return f"{self.__class__.__name__}(f={f_name}, *, h={v['h']})"


class RK4(Solver):
    A = _np.array([
        [0, 0, 0, 0],
        [1/2, 0, 0, 0],
        [0, 1/2, 0, 0],
        [0, 0, 1, 0]
    ])
    B = _np.array([1/6, 2/6, 2/6, 1/6])
    n_stages = 4


class Runge2(Solver):
    A = _np.array([
        [0, 0],
        [1/2, 0]
    ])
    B = _np.array([0, 1])
    n_stages = 2


def rk_step(f, t, u, A, B, K, h, **params):
    func = _partial(f, **params)
    K[0] = func(t, u)
    for s, a in enumerate(A[1:], start=1):
        c = sum(a[:s])
        K[s] = func(t + c*h, u + h*_np.dot(a[:s], (K[:s])))
    return u + h*_np.dot(K.T, B)


def solve_simple(f, t0, u0, n_loop, *, h=0.01, solver=RK4):
    sol = solver(f, h=h)
    u = _np.array(u0)
    t_seq = [t0 + i*h for i in range(n_loop)]

    u_seq = _np.array([u0, *[u := sol(t, u) for t in t_seq]])
    t_seq = _np.array([*t_seq, t0 + n_loop*h])
    return t_seq, u_seq



from functools import partial as _partial
from os import path as _path

from matplotlib import rcParams as _rcParams
import matplotlib.animation as _anm
import matplotlib.pyplot as _plt
import numpy as _np

import string as _string
from itertools import cycle as _cycle


config_drawing = {
    'figure.subplot.hspace':0.4,

    'lines.linewidth':0.8,

    'font.family':['Times New Roman', 'sans-serif'],
    'font.size':8,

    'xtick.direction':'in',
    'xtick.major.width':0.5,
    'xtick.major.size':2.5,
    'xtick.major.pad':7,
    'ytick.direction':'in',
    'ytick.major.width':0.5,
    'ytick.major.size':2.5,

    'ytick.right':True,
    'xtick.top':True,

    'axes.linewidth': 0.5,

    'axes.titlesize': 12,

    'axes.labelsize': 12,
    'mathtext.fontset':'cm',

    'scatter.marker':'.',

    'axes.labelpad':10,

    'legend.fancybox':False,
    'legend.edgecolor':'black',
    'legend.borderpad':0.5,
    'patch.linewidth':0.5,

    'grid.linewidth':0.5
    }


def _set_axis_label(ax, *xyz_labels, tex=True):
    for axis, label in zip(['x', 'y', 'z'], xyz_labels):
        method = getattr(ax, f'set_{axis}label')
        if tex:
            label = f'${label}$'
        method(label)


class Drawing(object):

    def __init__(self, rows=1, cols=1, number=False, three=False,
                 figsize=None, dpi=150, space=None, config=None,
                 number_place=None, number_size=None):

        if figsize is None:
            if rows <= cols:
                figsize = (3.14*1.7*2, 3.14*2)
            else:
                figsize = (3.14*2, 3.14*1.7*2)

        _rcParams.update(config or config_drawing)

        alphabets = _cycle(_string.ascii_lowercase)

        if three is True:
            three = tuple(range(1, rows*cols+1))
        elif isinstance(three, int):
            three = tuple([three])
        three = three or ()

        fig = _plt.figure(dpi=dpi, figsize=figsize)

        if three:
            _plt.subplots_adjust(wspace=0.4)

        if space is not None:
            _plt.subplots_adjust(wspace=space[0] ,hspace=space[1])



        axis = []
        for j in range(1, rows*cols+1):
            s = (rows, cols, j)
            kwargs = dict()

            if j in three:
                kwargs['projection']='3d'

            ax = fig.add_subplot(*s, **kwargs)
            axis.append(ax)

        axis = _np.array(axis).reshape((rows, cols))

        if number:
            x = 0.95 if number_place is None else number_place[0]
            y = 0.92 if number_place is None else number_place[1]
            size = (12 if number_size is None else number_size)
            for ax in _np.ravel(axis):
                ax.set_title(f'({next(alphabets)})', x=x, y=y, fontsize=size)

        for ax in axis.flatten():
            ax.set_axis_label =  _partial(_set_axis_label, ax)

        self.fig, self.ax = fig, axis

    def __getitem__(self,key):
        return self.ax[key]

    def cmap(self, cmap='viridis'):
        return _plt.cm.get_cmap(cmap)

    def legend(self, **params):
        _plt.legend(**params).get_frame().set_linewidth(0.5)

    def show(self):
        _plt.show()

    def save(self, path, dpi=200):
        _plt.savefig(path, dpi=dpi)

    def close(self):
        _plt.close()

    @classmethod
    def plot_a_and_b(cls, u_seq_a , u_seq_b=None, /, legend=True,
                     color=None, name=None, u_seq_more=[], *args, **kwargs):
        d = cls(1, 2, *args, **kwargs)
        ax_label = ['x', 'y', 'z']
        for i, (a, b) in enumerate(zip([0, 2], [1, 1])):
            color_cycle = _cycle(color or ['red', 'blue'])
            name_cycle = _cycle(name or ['a', 'b'])
            new_seq_list = ([u_seq_a] if u_seq_b is None
                            else [u_seq_a, u_seq_b, *u_seq_more])
            for u_seq in new_seq_list:
                c, n = next(color_cycle), next(name_cycle)
                d[0,i].plot(u_seq[:, a], u_seq[:, b], color=c, label=f'${n}$')
                d[0,i].scatter(u_seq[0, a], u_seq[0, b],
                               color=c, s=40, marker='o',
                               zorder=10, edgecolor='white')
                d[0,i].scatter(u_seq[-1, a], u_seq[-1, b],
                               color=c, s=40, marker='s',
                               zorder=10, edgecolor='white')

            d[0,i].set_axis_label(ax_label[a], ax_label[b])

        if legend:
            d[0, 0].legend()

        return d

    @classmethod
    def trajectory_3d(cls, u_seq, shadow=True, interval=30, *args, **kargs):

        x, y, z = u_seq[:, 0], u_seq[:, 1], u_seq[:, 2]
        u_left = _np.min(u_seq, axis=0)
        u_right = _np.max(u_seq, axis=0)
        scale = (u_right - u_left)/3
        u_left, u_right = u_left - scale, u_right + scale

        config_drawing['axes.labelsize']=20
        config_drawing['axes.labelpad']=0
        kargs['config']=config_drawing

        d = cls(1, 1, three=True, *args, **kargs)

        main_plot, = d[0,0].plot([], [], [], color='blue', linewidth=0.2)
        sub_plot, = d[0,0].plot([], [], [], color='red', linestyle="",
                                marker="o", markersize=1)

        if shadow:
            setting = {'linewidth':0.1, 'color': 'gray'}
            shadow_plot_1, = d[0,0].plot([], [],  [], **setting)
            shadow_plot_2, = d[0,0].plot([], [],  [], **setting)
            shadow_plot_3, = d[0,0].plot([], [],  [], **setting)

        def update(frame):
            s1 = slice(frame-1, frame+4)
            main_plot.set_data (x[0:frame], y[0:frame])
            main_plot.set_3d_properties(z[0:frame])

            sub_plot.set_data(x[s1], y[s1])
            sub_plot.set_3d_properties(z[s1])

            s = slice(0,frame+5)
            l = len(x[s])
            if shadow:
                plot_list = [shadow_plot_1, shadow_plot_2, shadow_plot_3]
                g_list = [(x[s], y[s], _np.full(l, u_left[2])),
                          (x[s],  _np.full(l, u_left[1]), z[s]),
                          (_np.full(l, u_right[0]), y[s], z[s])]
                for p, (x_s, y_s, z_s) in zip(plot_list, g_list):
                    p.set_data(x_s, y_s)
                    p.set_3d_properties(z_s)
                return main_plot, sub_plot, *plot_list

            else:
                return main_plot, sub_plot

        def init():
            d[0,0].set_xlim(u_left[0], u_right[0])
            d[0,0].set_ylim(u_left[1], u_right[1])
            d[0,0].set_zlim(u_left[2], u_right[2])
            d[0,0].xaxis.pane.set_facecolor("white")
            d[0,0].yaxis.pane.set_facecolor("white")
            d[0,0].zaxis.pane.set_facecolor("white")
            d[0,0].xaxis.pane.set_edgecolor('black')
            d[0,0].yaxis.pane.set_edgecolor('black')
            d[0,0].zaxis.pane.set_edgecolor('black')
            d[0,0].xaxis.line.set_linewidth(0.2)
            d[0,0].yaxis.line.set_linewidth(0.2)
            d[0,0].zaxis.line.set_linewidth(0.2)

            d[0,0].set_axis_label('x', 'y', 'z')
            d[0,0].grid(False)
            d[0,0].set_xticks([])
            d[0,0].set_yticks([])
            d[0,0].set_zticks([])

            d[0,0].view_init(azim=127,elev=27)

            return main_plot,

        init()
        ani = _anm.FuncAnimation(d.fig, func=update, init_func=init, blit=True,
                                 interval=interval, save_count=len(x))

        return d, ani



# Mutual Information

import numpy as _np

from ._utils import (reshape as _reshape, get_bottom as _get_bottom)
from ..utils import Drawing as _Drawing

def _log2(a):
    return _np.log2(a+1e-15)


def calc_mutual_info(x, y, N, **options):
    H_x = calc_entropy(x, N, **options)
    H_y = calc_entropy(y, N, **options)
    H_xy = calc_joint_entropy(x, y, N, **options)
    return H_x + H_y - H_xy


def calc_entropy(x, N, **options):
    hist, _ = _np.histogram(x, **options)
    ps = hist/N
    H_x = -_np.sum(ps*_log2(ps))
    return H_x


def calc_joint_entropy(x, y, N, **options):
    hist, _, _ = _np.histogram2d(x, y, **options)
    ps = (hist/N).flatten()
    H_xy = -_np.sum(ps*_log2(ps))
    return H_xy


def mutual_info(u_seq, tau):
    u_seq = _reshape(u_seq)

    N, dim = u_seq.shape

    miss = []
    for i in range(dim):
        us = u_seq[:, i]
        mis = [calc_mutual_info(us[:-t], us[t:], N, bins=32)
               for t in range(1, tau+1)]
        miss.append(mis)

    return _np.array(miss).T


def est_lag_w_mi(u_seq, tau_max, plot=True, path_save_plot=None):
    mi_seq = mutual_info(u_seq, tau_max)
    bottom = _get_bottom(mi_seq)

    if plot:
        d = _Drawing()
        for i, idx in enumerate(bottom):
            d[0,0].plot(mi_seq[:, i], label=f'{i}',
                        marker='.', markersize=5, zorder=5)
            d[0,0].scatter(idx, mi_seq[idx, i], s=70, color='red', zorder=10)

        d[0,0].set_axis_label('Time~lag', 'Mutual~Information~[bit]')
        d[0,0].set_xlim(0, tau_max-1)
        d[0,0].legend()
        if path_save_plot is not None:
            d.save(path_save_plot)
        d.show()
        d.close()

    return list(bottom)



from math import dist as _dist
from itertools import combinations as _combinations

import numpy as _np

from ..utils._draw import Drawing as _Drawing


def simple_threshold(ds, theta=0.5):
    if (d_max:=_np.max(ds))!=0:
        pv = (ds/d_max>theta)*255
        return pv
    return ds


def calc_recurrence_plot(u_seq, rule=simple_threshold, *params, **kargs):
    size = len(u_seq)
    rp = _np.zeros((size,size))

    ds =  _np.array([_dist(u1, u2) for u1, u2 in _combinations(u_seq, 2)])

    pv = rule(ds, *params, **kargs)

    rp[_np.triu(_np.ones((size,size), dtype=bool), k=1)] = pv

    return rp+rp.T


def show_recurrence_plot(u_seq, rule=simple_threshold, cmap=False,
                         path_save_plot=None,
                         *params, **kargs):
    rp = calc_recurrence_plot(u_seq, rule, *params, **kargs)

    d = _Drawing()
    im = d[0,0].imshow(rp, origin='lower')
    if cmap:
        d.fig.colorbar(im)

    if path_save_plot is not None:
        d.save(path_save_plot)

    d.show()
    d.close()

    return rp



# False Nearest Neighbors - Algorithm

import warnings as _warnings

import numpy as _np
from scipy.spatial.distance import cdist as _cdist

from ._utils import embedding as _embedding
from ..utils import Drawing as _Drawing

def _dist(seq):
    return _cdist(seq, seq, metric='euclidean')


def fnn(u_seq, threshold_R=10, threshold_A=2, T=50, D_max=10):
    msg = 'It will not be available after version 0.2. ' \
          'Use `est_dimension_w_fnn` instead.'
    _warnings.warn(msg)

    est_dimension_w_fnn(u_seq, T=T, D_max=D_max,
                        threshold_R=threshold_R, threshold_A=threshold_A)


def est_dimension_w_fnn(u_seq, T, D_max=10,
                        threshold_R=10, threshold_A=2, threshold_percent=1,
                        plot=True, path_save_plot=None):

    R_A = _np.std(u_seq)

    e_seq_list = [_embedding(u_seq, T, j) for j in range(1, D_max+2)]

    percentage_list = []
    for e_seq1, e_seq2 in zip(e_seq_list, e_seq_list[1:]):
        dist1 = _dist(e_seq1) + _np.eye(len(e_seq1))*10000
        dist2 = _dist(e_seq2)

        idx_n_list = dist1[:(l2 := len(dist2)), :l2].argmin(axis=0)

        dist1_min = dist1.min(axis=0)
        dist2_min = _np.array([d[idx] for idx, d in zip(idx_n_list, dist2)])

        percentage = 0
        for R_Dk1, R_Dk2 in zip(dist1_min, dist2_min):
            d = _np.sqrt((R_Dk2**2-R_Dk1**2)/(R_Dk1**2))
            criterion_1 = d > threshold_R
            criterion_2 = R_Dk2/R_A >= threshold_A
            if criterion_1 and criterion_2:
                percentage += 1
        percentage_list.append(percentage*(1/len(e_seq2)))

    percentages = _np.array(percentage_list)*100
    dranges = _np.arange(1, len(percentages)+1)
    dimension = dranges[percentages<threshold_percent][0]

    if plot:
        d = _Drawing()
        d[0,0].plot(dranges, percentages,
                    marker='.', markersize=5, color='tab:blue', zorder=5)
        d[0,0].axhline(threshold_percent,
                       color='black', linestyle='dashed', linewidth=0.5,
                       zorder=0)
        d[0,0].scatter(dimension, percentages[dimension-1],
                       s=70, color='red', zorder=10)

        d[0,0].set_axis_label('Embedding~Dimension', r'False~NN~Percentage~[\%]')
        if path_save_plot is not None:
            d.save(path_save_plot)
        d.show()

    return dimension, percentages



from ..systems._differential import Differential as _Differential


class Lorenz(_Differential):

    def parameter(self, s=10, r=28, b=8/3):
        self.s, self.r, self.b = s, r, b
        self.dim = 3

    def equation(self, t, u):
        s, r, b = self.s, self.r, self.b

        x, y, z = u

        x_dot = s*(y - x)
        y_dot = r*x - y - x*z
        z_dot = x*y - b*z

        return x_dot, y_dot, z_dot

    def jacobian(self):
        s, r, b = self.s, self.r, self.b
        x, y, z = self.u

        j = [[-s, s, 0],
             [r-z, -1, -x],
             [y, x, -b]]

        return j



from itertools import accumulate as _accumulate
from operator import add as _add

import numpy as _np


def calc_les_difference(difference, N=5000, **options):
    model = _make_model(difference, **options)

    if model.dim == 1:
        return calc_les_difference_dim_1(model, N)
    else:
        return calc_les_difference_dim_geq_2(model, N)


def calc_les_difference_dim_geq_2(difference, N, n_average=100):
    if difference.inf:
        raise ValueError('Initial value is outside basin of attraction')

    Q, R = _np.linalg.qr(difference.j())

    R_list= [_np.diag(R)]
    les = [_np.log(_np.abs(_np.diag(R)))]
    for _ in range(N):
        difference.solve(*difference.internal_state)
        J = difference.j()
        Q, R = _np.linalg.qr(J@Q)
        R_list.append(_np.diag(R))
        les.append(_np.log(_np.abs(_np.diag(R))))

    les_list = [l/i for i, l in enumerate(_accumulate(les, _add), 1)]

    return _np.array(les_list), _np.average(les_list[-n_average:], axis=0)


def calc_les_difference_dim_1(difference, N, n_average=100):
    le_list = []
    for _ in range(N):
        difference.solve(*difference.internal_state)
        le_list.append(_np.log(_np.abs(difference.j())))
    le_list = _np.array(
        [le/i for i, le in enumerate(_accumulate(le_list, _add), 1) ])
    return le_list, _np.array([_np.average(le_list[-n_average:])])

def _make_model(difference, u0=None, **options):
    if u0 is None:
        model = difference.on_attractor(u0=u0, **options)
    else:
        model = difference(**options)
        model.u, model.t = u0, 0
    return model



from ._afn import afn, est_dimension_w_afn
from ._autocorrelation import acf, est_lag_w_acf
from ._fnn import fnn, est_dimension_w_fnn
from ._gp import calc_correlation_dimention_w_gp
from ._mutualinfo import mutual_info, est_lag_w_mi
from ._recurrenceplot import (calc_recurrence_plot, show_recurrence_plot,
                              simple_threshold)
from ._utils import embedding, get_bottom, bartlett, get_minidx_below_seq
from ._wayland import wayland, est_dimension_w_wayland
from ._box3 import (calc_dimension_capacity, calc_dimension_information,
                    calc_dimension_correlation)


# 力学系(dynamical systems)

from abc import ABC as _ABC, abstractmethod as _abstractmethod

import numpy as _np

from ._tu import TU as _TU


class DynamicalSystems(_ABC):

    def __init__(self, t=None, u=None, **params):
        self.dim = 0
        self.parameter()

        if self.dim == 0:
            class_name = self.__class__.__name__
            msg = (f"need to set {class_name}'s dimension "
                   f"({class_name}.dim=? in parameter())")
            raise NotImplementedError(msg)

        self.t = t or 0
        self.u = u if u is not None else _np.zeros(self.dim)

        self._t_seq, self._u_seq = [], []

        self.parameter(**params)

    @property
    def inf(self):
        return any(_np.isinf(self.u))

    @property
    def internal_state(self):
        return _TU(self.t, self.u)

    @property
    def t_seq(self):
        return _np.array(self._t_seq)

    @property
    def u_seq(self):
        return _np.array(self._u_seq)

    @classmethod
    def on_attractor(cls, t0=None, u0=None, h=0.01, *, T_0=5000, **params):
        c = cls(t0, u0)
        c.parameter(**params)
        c.settle_on_attractor(t0, u0, h=h, T_0=T_0)
        return c

    @classmethod
    def get_u_seq(cls, n, *args, **kwargs):
        c = cls.on_attractor(*args, **kwargs)
        c.solve_n_times(n)
        return c.u_seq

    @_abstractmethod
    def equation(self, t, u):
        """equation"""

    def j(self, **params):
        return _np.array(self.jacobian(**params))

    def jacobian(self):
        """jacobian"""
        return None

    def make_inital(self):
        return _np.random.rand(self.dim)

    def parameter(self):
        """set parameter for equation"""

    def reset_u_seq(self):
        self._u_seq = []

    def settle_on_attractor(self, t0=None, u0=None,
                            *, T_0=5000, notsave=True, **params):
        self.u = self.make_inital() if u0 is None else u0
        self.t = t0 or 0

        for _ in range(T_0):
            self.solve(*self.internal_state, **params)

        if notsave:
            self._u_seq, self._t_seq = [], []

        if t0 is None:
            self.t = 0

        self.t, self.u = self.internal_state

        return self.internal_state

    def solve(self, *args, **kwargs):
        tu = self._solve(*args, **kwargs)

        if kwargs.get('save', True):
            self._u_seq.append(tu.u)
            self._t_seq.append(tu.t)

        return tu

    def solve_n_times(self, n):
        for _ in range(n):
            self.solve(*self.internal_state)
        return self.t_seq, self.u_seq

    def __call__(self, t, u):
        return _np.array(self.equation(t, u))

    def __repr__(self):
        v = vars(self)

        p = ', '.join(f'{key}={_np.round(v[key], 3)}'
                      for key in v.keys()
                      if ('_' not in key) and (key not in ['t', 'u']))

        name = self.__class__.__name__
        return f'{name}({p})'

    @_abstractmethod
    def _solve(self, t, u):
        self.t, self.u = t, u
        return _TU(self.t, self.u)



from itertools import accumulate as _accumulate
from operator import add as _add

import numpy as _np
from scipy import signal as _signal
from scipy.stats import norm as _norm

def reshape(u_seq):
    if (ndim:=u_seq.ndim) == 1:
        u_seq = u_seq.reshape(len(u_seq), 1)
    elif ndim >= 3:
        raise NotImplementedError(f'{ndim} ndim is not supported')
    return u_seq


def embedding_seq(u_seq, T, D):
    idx = _np.arange(0,D,1)*T
    return _np.array([u_seq[idx+i,:] for i in range(len(u_seq)-(D-1)*T)])


def embedding_seq_1dim(u_seq, T, D):
    idx = _np.arange(0,D,1)*T
    e_seq = _np.array([u_seq[idx+i,:] for i in range(len(u_seq)-(D-1)*T)])
    return e_seq.reshape(len(e_seq), D)


def embedding(u_seq, T, D):
    dim, length = u_seq.ndim, len(u_seq)
    if len(u_seq.shape)==1:
        u_seq = u_seq.reshape(length, dim)

    idx = _np.arange(0,D,1)*T
    e_seq = _np.array([u_seq[idx+i,:] for i in range(length-(D-1)*T)])

    if u_seq.shape[1] == 1:
        e_seq = e_seq.reshape(len(e_seq), D)
    return e_seq


def get_bottom(seq, threshold=float('inf')):
    _, dim = seq.shape

    lags = []
    for i in range(dim):
        us = seq[:, i]
        min_idx = _signal.argrelmin(us, order=1)[0]
        candidate = min_idx[us[min_idx]<=threshold]
        if len(candidate):
            lags.append(candidate[0])
        else:
            lags.append(None)
    return tuple(lags)


def bartlett(seq, alpha=0.95):
    _, dim = seq.shape
    N = len(seq)
    var = [_np.ones(dim)/N,
           *[(1+2*rho)/N for rho in _accumulate(seq[1:,]**2, _add)]]
    z = _norm.ppf(alpha)
    return z*_np.array(var)**(1/2)


def get_minidx_below_seq(rho_seq, var_seq):
    _, dim = rho_seq.shape
    return [(rho_seq[:, i]<var_seq[:, i]).argmax() for i in range(dim)]



# Averaged False Neighbors - Algorithm

import warnings as _warnings

import numpy as _np
from scipy.spatial.distance import cdist as _cdist

from ._utils import embedding as _embedding
from ..utils import Drawing as _Drawing


def _dist(seq):
    return _cdist(seq, seq, metric='chebyshev')


def afn(u_seq, T=1, D_max=10):
    msg = 'It will not be available after version 0.2. ' \
          'Use `est_dimension_w_afn` instead.'
    _warnings.warn(msg)
    return est_dimension_w_afn(u_seq, T=T, D_max=D_max)[1]


def est_dimension_w_afn(u_seq, T, D_max=10,
                        threshold_E1=0.9, threshold_E2=1,
                        plot=True, path_save_plot=None):

    R_A = _np.std(u_seq)

    e_seq_list = [_embedding(u_seq, T, j) for j in range(1, D_max+3)]

    a_list, b_list = [], []
    for e_seq1, e_seq2 in zip(e_seq_list, e_seq_list[1:]):
        dist1 = _dist(e_seq1) + _np.eye(len(e_seq1))*10000
        dist2 = _dist(e_seq2)

        idx_n_list = dist1[:(l2 := len(dist2)), :l2].argmin(axis=0)

        dist1_min = dist1.min(axis=0)
        dist2_min = _np.array([d[idx] for idx, d in zip(idx_n_list, dist2)])

        # E2の計算のためのEの計算
        b = [_np.abs(e_line[-1] - e_seq2[idx][-1])/R_A
             for idx, e_line in zip(idx_n_list, e_seq2)]
        b_bar = _np.average(b)
        b_list.append(b_bar)

        # E1の計算のためのEの計算
        a = [R_Dk2/R_Dk1 for R_Dk1, R_Dk2 in zip(dist1_min, dist2_min)]
        a_bar = _np.average(a)
        a_list.append(a_bar)

    E_list = []
    for c in [a_list, b_list]:
        criterion = []
        for tau, tau_p1 in zip(c, c[1:]):
            criterion.append(tau_p1/tau)
        E_list.append(_np.array(criterion))

    dranges = _np.arange(1, len(E_list[0])+1)

    dimension_list = []
    for E, threshold in zip(E_list, [threshold_E1, threshold_E2]):
        if len(rule:=(E>threshold)) > 0:
            dim = dranges[rule][0]
        else:
            dim = None
        dimension_list.append(dim)

    if plot:
        d = _Drawing()
        for E, threshold, label, dim in zip(E_list,
                                            [threshold_E1, threshold_E2],
                                            ['E1', 'E2'],
                                            dimension_list):
            if dim is not None:
                d[0,0].scatter(dim, E[dim-1],
                               s=70, color='red', zorder=10)

            d[0,0].plot(dranges, E, label=label,
                        marker='.', markersize=5, zorder=5)

            d[0,0].axhline(threshold,
                           color='black', linestyle='dashed', linewidth=0.5,
                           zorder=0)
        d[0,0].legend()
        d[0,0].set_axis_label('Embedding~Dimension', r'E1~&~E2')
        if path_save_plot is not None:
            d.save(path_save_plot)
        d.show()

    return dimension_list, E_list



# Wayland - Algorithm

from random import sample as _sample
from statistics import median as _median

import numpy as _np

from ._utils import (embedding_seq as _embedding_seq,
                     reshape as _reshape)
from ..utils import Drawing as _Drawing


def wayland(u_seq, K=4, tau=5, Q=10, M = 51, D_max=10, reshapemode=True):
    _, median_e_trans_ave= est_dimension_w_wayland(
        u_seq=u_seq, T=1, D_max=D_max,
        tau=tau, K=K, Q=Q, M=51, reshape=reshapemode,
        plot=False)
    return median_e_trans_ave


def est_dimension_w_wayland(u_seq, T, D_max=10,
                            tau=5, K=4, Q=10, M=51,
                            reshape=True,
                            plot=True, path_save_plot=None):
    if reshape:
        u_seq = _reshape(u_seq)

    median_e_trans_ave = []
    for D in range(1, D_max+1):
        e_seq = _embedding_seq(u_seq, T, D)
        N = len(e_seq)
        median_e_trans = []
        for _ in range(Q):
            e_trans = []
            for idx_n in _sample(range(len(e_seq)), M):
                e_n = e_seq[idx_n]

                idx_distance = ((i, _np.linalg.norm(e))
                                for i, e in enumerate(e_seq[:N-tau]-e_n))
                idx_distance = sorted(idx_distance, key=lambda x: x[1])[1:K+1]

                v_seq = _np.array([
                    e_seq[idx + tau] - e_seq[idx] for idx, _ in idx_distance])
                v_bar = _np.sum(v_seq, axis=0)/(K+1)
                bunshi = [_np.linalg.norm(v)/_np.linalg.norm(v_bar)
                          for v in v_seq - v_bar]
                e_trans.append(sum(bunshi)/(K+1))

            median_e_trans.append(_median(e_trans))
        median_e_trans_ave.append(_np.average(median_e_trans))

    median_e_trans_ave = _np.array(median_e_trans_ave)
    dimension = median_e_trans_ave.argmin()+1

    if plot:
        d = _Drawing()
        d[0,0].plot(range(1, len(median_e_trans_ave)+1), median_e_trans_ave,
                    marker='.', markersize=5, color='tab:blue', zorder=5)
        d[0,0].scatter(median_e_trans_ave.argmin()+1, median_e_trans_ave.min(),
                        s=70, color='red', zorder=10)
        d[0,0].set_axis_label('Dimension', 'median(E_{trans})')
        d[0,0].set_yscale('log')
        if path_save_plot is not None:
            d.save(path_save_plot)
        d.show()

    return dimension, median_e_trans_ave


if __name__ == '__main__':
    from ..equations import Lorenz

    u_seq = Lorenz.get_u_seq(5000)[::5, 0]

    est_dimension_w_wayland(u_seq, T=1)



import os
from setuptools import setup, find_packages
import sys


if sys.version_info[:2] < (3, 8):
    raise RuntimeError('Python version >= 3.8 required.')


def _requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


NAME = 'hundun'

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), (version := {}))

with open('README.md') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=version['__version__'],
    description='hundun is a python library for the exploration of chaos.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/llbxg/hundun',
    author='kosh',
    author_email='llllbxg@gmial.com',
    license='MIT',
    packages=find_packages(),
    install_requires=_requirements(),
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.8',
    classifiers=[
        'Topic :: Scientific/Engineering :: Physics'
    ]
    )



from ._func import calc_lyapunov_dimension
from ._les_differential import (calc_les_differential,
                                calc_les_differential_w_qr,
                                calc_max_le_differential,
                                calc_les_differential_w_orth)
from ._les_difference import calc_les_difference


# def

import numpy as _np

from ..systems._systems import DynamicalSystems as _DynamicalSystems
from ..systems._differential import Differential as _Differential
from ..systems._difference import Difference as _Difference


def calc_les(system, **options):

    if issubclass(system, _Differential):
        return calc_les_differential(system, **options)

    elif issubclass(system, _Difference):
        return calc_les_difference(system, **options)

    if not issubclass(system, _DynamicalSystems):
        raise TypeError(f"{system} must be a DynamicalSystems")

    return _np.zeros((3, 3)), (0, 0, 0)
